from os import mkdir
from os.path import expanduser, join
from datetime import datetime
from logging import getLogger, FileHandler, StreamHandler

from cml.shared.errors import NoModelReconstructedError
from cml.shared.parameter import PROTOCOL_LEVEL
from cml.usecases.usecase import Usecase
from cml.domain import (
    Preprocessor,
    DataSource,
    LearnblockIdentifier,
    Constructor,
    FeatureSelector,
    Reconstructor
)


class PreprocessUsecase(Usecase):
    def process(self, request):
        preprocesser = Preprocessor(request.pre_settings)
        cleaned_data = preprocesser.clean(request.source_adapter)
        learnblock_identifier = LearnblockIdentifier(request.block_settings,
                                                     request.density_estimator,
                                                     request.relative_extrema)
        data_source = DataSource(cleaned_data,
                                 learnblock_identifier,
                                 block_size=request.block_settings.block_size)
        return data_source


class CreateConstructorUsecase(Usecase):
    def process(self, request):
        constructor = Constructor(request.ml_models, request.settings)
        constructor.construction_type = request.settings.construct_type
        return constructor


class FeatureSelectionUsecase(Usecase):
    def process(self, request):
        return FeatureSelector(request.filter_method,
                               request.embedded_method,
                               request.settings)


class ReconstructionUsecase(Usecase):
    def process(self, request):
        return Reconstructor(request.settings,
                             request.ml_models,
                             request.knowledge_domain)


class KnowledgeSearchUsecase(Usecase):
    def process(self, request):
        source, constr, f_selectr, reconstr = self._init_loggers(request)
        reconstr.category = constr.construction_type

        for learnblock in source.learnblocks:
            for complete_learnblock in constr.construct(learnblock):
                try:
                    reduced_learnblock = f_selectr.select(complete_learnblock)
                    pragmatic_model = reconstr.reconstruct(reduced_learnblock)
                except ValueError as error:
                    # Can't reduce learnblock
                    # log
                    continue
                except NoModelReconstructedError:
                    continue
                else:
                    # process with reconstruction
                    if pragmatic_model:
                        print(pragmatic_model.model.mean_error)

    def _init_loggers(self, request):
        log_dir = create_log_dir()
        iteration_log_file = join(log_dir, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")
        iteration_logger.addHandler(iteration_handler)
        request.data_source.logger = iteration_logger
        request.constructor.logger = iteration_logger
        request.feature_selector.logger = iteration_logger

        if request.stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)

        return (request.data_source,
                request.constructor,
                request.feature_selector,
                request.reconstructor)


def create_log_dir():
    log_dir = join(expanduser('~'),
                   ".cml",
                   datetime.now().strftime("%d%m%Y%H%M%S"))
    mkdir(log_dir)
    return log_dir

